__author__ = "Duncan Johnstone"
__copyright__ = "Copyright 2017-2019, Python Crystallographic Electron Microscopy"
__credits__ = ["Duncan Johnstone", "Phillip Crout", "Ben Martineau", "Simon Hogas"]
__license__ = "GPL"
__version__ = "0.8.1"
__maintainer__ = "Duncan Johnstone"
__email__ = "dnj23@cam.ac.uk"
__status__ = "Development"
